from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.ops import Op
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_function
from sonolus.script.quad import QuadLike, flatten_quad
from sonolus.script.record import Record


class Particle(Record):
    """A particle effect."""

    id: int

    @property
    def is_available(self) -> bool:
        """Check if the particle effect is available."""
        return _has_particle_effect(self.id)

    def spawn(self, quad: QuadLike, duration: float, loop: bool = False) -> ParticleHandle:
        """Spawn the particle effect.

        Args:
            quad: The quad to spawn the particle effect on.
            duration: The duration of the particle effect.
            loop: Whether to loop the particle effect.

        Returns:
            ParticleHandle: A handle to the spawned particle effect.
        """
        return ParticleHandle(_spawn_particle_effect(self.id, *flatten_quad(quad), duration, loop))


class ParticleHandle(Record):
    """A handle to a looping particle effect."""

    id: int

    def move(self, quad: QuadLike) -> None:
        """Move the particle effect to a new location.

        Args:
            quad: The new quad to move the particle effect to.
        """
        _move_particle_effect(self.id, *flatten_quad(quad))

    def destroy(self) -> None:
        """Destroy the particle effect."""
        _destroy_particle_effect(self.id)


@native_function(Op.HasParticleEffect)
def _has_particle_effect(particle_id: int) -> bool:
    raise NotImplementedError


@native_function(Op.SpawnParticleEffect)
def _spawn_particle_effect(
    particle_id: int,
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    x4: float,
    y4: float,
    duration: float,
    loop: bool,
) -> int:
    raise NotImplementedError


@native_function(Op.MoveParticleEffect)
def _move_particle_effect(
    handle: int, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float
) -> None:
    raise NotImplementedError


@native_function(Op.DestroyParticleEffect)
def _destroy_particle_effect(handle: int) -> None:
    raise NotImplementedError


@dataclass
class _ParticleInfo:
    name: str


def particle(name: str) -> Any:
    """Define a particle with the given name."""
    return _ParticleInfo(name)


type Particles = NewType("Particles", Any)  # type: ignore


@dataclass_transform(kw_only_default=True)
def particles[T](cls: type[T]) -> T | Particles:
    """Decorator to define particles.

    Usage:
        ```python
        @particles
        class Particles:
            tap: StandardParticle.NOTE_CIRCULAR_TAP_RED
            other: Particle = particle("other")
        ```
    """
    if len(cls.__bases__) != 1:
        raise ValueError("Particles class must not inherit from any class (except object)")
    instance = cls()
    names = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for particles: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if annotation_type is not Particle:
            raise TypeError(f"Invalid annotation for particles: {annotation}, expected annotation of type Particle")
        if len(annotation_values) != 1 or not isinstance(annotation_values[0], _ParticleInfo):
            raise TypeError(
                f"Invalid annotation for particles: {annotation}, expected a single string annotation value"
            )
        particle_name = annotation_values[0].name
        names.append(particle_name)
        setattr(instance, name, Particle(i))
    instance._particles_ = names
    instance._is_comptime_value_ = True
    return instance


class StandardParticle:
    """Standard particles."""

    NOTE_CIRCULAR_TAP_NEUTRAL = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_NEUTRAL")]
    NOTE_CIRCULAR_TAP_RED = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_RED")]
    NOTE_CIRCULAR_TAP_GREEN = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_GREEN")]
    NOTE_CIRCULAR_TAP_BLUE = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_BLUE")]
    NOTE_CIRCULAR_TAP_YELLOW = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_YELLOW")]
    NOTE_CIRCULAR_TAP_PURPLE = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_PURPLE")]
    NOTE_CIRCULAR_TAP_CYAN = Annotated[Particle, particle("#NOTE_CIRCULAR_TAP_CYAN")]
    NOTE_CIRCULAR_ALTERNATIVE_NEUTRAL = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_NEUTRAL")]
    NOTE_CIRCULAR_ALTERNATIVE_RED = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_RED")]
    NOTE_CIRCULAR_ALTERNATIVE_GREEN = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_GREEN")]
    NOTE_CIRCULAR_ALTERNATIVE_BLUE = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_BLUE")]
    NOTE_CIRCULAR_ALTERNATIVE_YELLOW = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_YELLOW")]
    NOTE_CIRCULAR_ALTERNATIVE_PURPLE = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_PURPLE")]
    NOTE_CIRCULAR_ALTERNATIVE_CYAN = Annotated[Particle, particle("#NOTE_CIRCULAR_ALTERNATIVE_CYAN")]
    NOTE_CIRCULAR_HOLD_NEUTRAL = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_NEUTRAL")]
    NOTE_CIRCULAR_HOLD_RED = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_RED")]
    NOTE_CIRCULAR_HOLD_GREEN = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_GREEN")]
    NOTE_CIRCULAR_HOLD_BLUE = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_BLUE")]
    NOTE_CIRCULAR_HOLD_YELLOW = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_YELLOW")]
    NOTE_CIRCULAR_HOLD_PURPLE = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_PURPLE")]
    NOTE_CIRCULAR_HOLD_CYAN = Annotated[Particle, particle("#NOTE_CIRCULAR_HOLD_CYAN")]
    NOTE_LINEAR_TAP_NEUTRAL = Annotated[Particle, particle("#NOTE_LINEAR_TAP_NEUTRAL")]
    NOTE_LINEAR_TAP_RED = Annotated[Particle, particle("#NOTE_LINEAR_TAP_RED")]
    NOTE_LINEAR_TAP_GREEN = Annotated[Particle, particle("#NOTE_LINEAR_TAP_GREEN")]
    NOTE_LINEAR_TAP_BLUE = Annotated[Particle, particle("#NOTE_LINEAR_TAP_BLUE")]
    NOTE_LINEAR_TAP_YELLOW = Annotated[Particle, particle("#NOTE_LINEAR_TAP_YELLOW")]
    NOTE_LINEAR_TAP_PURPLE = Annotated[Particle, particle("#NOTE_LINEAR_TAP_PURPLE")]
    NOTE_LINEAR_TAP_CYAN = Annotated[Particle, particle("#NOTE_LINEAR_TAP_CYAN")]
    NOTE_LINEAR_ALTERNATIVE_NEUTRAL = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_NEUTRAL")]
    NOTE_LINEAR_ALTERNATIVE_RED = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_RED")]
    NOTE_LINEAR_ALTERNATIVE_GREEN = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_GREEN")]
    NOTE_LINEAR_ALTERNATIVE_BLUE = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_BLUE")]
    NOTE_LINEAR_ALTERNATIVE_YELLOW = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_YELLOW")]
    NOTE_LINEAR_ALTERNATIVE_PURPLE = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_PURPLE")]
    NOTE_LINEAR_ALTERNATIVE_CYAN = Annotated[Particle, particle("#NOTE_LINEAR_ALTERNATIVE_CYAN")]
    NOTE_LINEAR_HOLD_NEUTRAL = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_NEUTRAL")]
    NOTE_LINEAR_HOLD_RED = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_RED")]
    NOTE_LINEAR_HOLD_GREEN = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_GREEN")]
    NOTE_LINEAR_HOLD_BLUE = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_BLUE")]
    NOTE_LINEAR_HOLD_YELLOW = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_YELLOW")]
    NOTE_LINEAR_HOLD_PURPLE = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_PURPLE")]
    NOTE_LINEAR_HOLD_CYAN = Annotated[Particle, particle("#NOTE_LINEAR_HOLD_CYAN")]
    LANE_CIRCULAR = Annotated[Particle, particle("#LANE_CIRCULAR")]
    LANE_LINEAR = Annotated[Particle, particle("#LANE_LINEAR")]
    SLOT_CIRCULAR = Annotated[Particle, particle("#SLOT_CIRCULAR")]
    SLOT_LINEAR = Annotated[Particle, particle("#SLOT_LINEAR")]
    JUDGE_LINE_CIRCULAR = Annotated[Particle, particle("#JUDGE_LINE_CIRCULAR")]
    JUDGE_LINE_LINEAR = Annotated[Particle, particle("#JUDGE_LINE_LINEAR")]


@particles
class EmptyParticles:
    pass



from collections.abc import Callable
from typing import Literal

import pytest

from sonolus.backend.finalize import cfg_to_engine_node
from sonolus.backend.mode import Mode
from sonolus.backend.node import format_engine_node
from sonolus.backend.optimize.flow import cfg_to_text
from sonolus.backend.optimize.passes import OptimizerConfig, run_passes
from sonolus.build.compile import callback_to_cfg
from sonolus.build.engine import package_engine
from sonolus.script.archetype import _BaseArchetype
from sonolus.script.internal.callbacks import (
    CallbackInfo,
    navigate_callback,
    preprocess_callback,
    update_callback,
    update_spawn_callback,
)
from sonolus.script.internal.context import GlobalContextState, ReadOnlyMemory
from sonolus.script.project import BuildConfig
from tests.regressions import pydori_project
from tests.regressions.conftest import compare_with_reference

PROJECTS = {
    "pydori": pydori_project,
}

PASSES = {
    # minimal won't work since some callbacks will run out of temporary memory
    "fast": BuildConfig.FAST_PASSES,
    "standard": BuildConfig.STANDARD_PASSES,
}


@pytest.mark.parametrize("project", ["pydori"])
@pytest.mark.parametrize("passes", ["fast", "standard"])
def test_project_full_build_succeeds(
    project: Literal["pydori"],
    passes: Literal["fast", "standard"],
):
    package_engine(
        PROJECTS[project].engine.data,
        BuildConfig(
            passes=PASSES[passes],
        ),
    )


@pytest.mark.parametrize("project", ["pydori"])
@pytest.mark.parametrize("passes", ["fast", "standard"])
def test_project_method_build_regressions(
    project: Literal["pydori"],
    passes: Literal["fast", "standard"],
):
    engine = PROJECTS[project].engine.data

    _build_mode_callbacks(
        project_name=project,
        mode=Mode.PLAY,
        archetypes=engine.play.archetypes,
        global_callbacks=None,
        passes=passes,
    )
    _build_mode_callbacks(
        project_name=project,
        mode=Mode.WATCH,
        archetypes=engine.watch.archetypes,
        global_callbacks=[(update_spawn_callback, engine.watch.update_spawn)],
        passes=passes,
    )
    _build_mode_callbacks(
        project_name=project,
        mode=Mode.PREVIEW,
        archetypes=engine.preview.archetypes,
        global_callbacks=None,
        passes=passes,
    )
    _build_mode_callbacks(
        project_name=project,
        mode=Mode.TUTORIAL,
        archetypes=None,
        global_callbacks=[
            (preprocess_callback, engine.tutorial.preprocess),
            (navigate_callback, engine.tutorial.navigate),
            (update_callback, engine.tutorial.update),
        ],
        passes=passes,
    )


def _build_mode_callbacks(
    project_name: str,
    mode: Mode,
    archetypes: list[type[_BaseArchetype]] | None,
    global_callbacks: list[tuple[CallbackInfo, Callable]] | None,
    passes: str,
):
    for archetype in archetypes or []:
        archetype._init_fields()

        callback_items = [
            (cb_name, cb_info, getattr(archetype, cb_name))
            for cb_name, cb_info in archetype._supported_callbacks_.items()
            if getattr(archetype, cb_name) not in archetype._default_callbacks_
        ]

        for cb_name, cb_info, cb in callback_items:
            global_state = GlobalContextState(
                mode,
                {a: i for i, a in enumerate(archetypes)} if archetypes is not None else None,
                ReadOnlyMemory(),
            )
            cfg = callback_to_cfg(global_state, cb, cb_info.name, archetype)
            compare_with_reference(
                f"{project_name}_{mode.name.lower()}_{camel_to_snake(archetype.__name__)}_{cb_name}_cfg",
                cfg_to_text(cfg),
            )
            cfg = run_passes(cfg, PASSES[passes], OptimizerConfig(mode=mode, callback=cb_info.name))
            compare_with_reference(
                f"{project_name}_{mode.name.lower()}_{camel_to_snake(archetype.__name__)}_{cb_name}_{passes}_optimized_cfg",
                cfg_to_text(cfg),
            )
            node = cfg_to_engine_node(cfg)
            compare_with_reference(
                f"{project_name}_{mode.name.lower()}_{camel_to_snake(archetype.__name__)}_{cb_name}_{passes}_nodes",
                format_engine_node(node),
            )

    for cb_info, cb in global_callbacks or []:
        global_state = GlobalContextState(
            mode,
            {a: i for i, a in enumerate(archetypes)} if archetypes is not None else None,
            ReadOnlyMemory(),
        )
        cfg = callback_to_cfg(global_state, cb, cb_info.name, None)
        compare_with_reference(
            f"{project_name}_{mode.name.lower()}_global_{camel_to_snake(cb_info.name)}_{passes}_cfg",
            cfg_to_text(cfg),
        )
        cfg = run_passes(cfg, PASSES[passes], OptimizerConfig(mode=mode, callback=cb_info.name))
        compare_with_reference(
            f"{project_name}_{mode.name.lower()}_global_{camel_to_snake(cb_info.name)}_{passes}_optimized_cfg",
            cfg_to_text(cfg),
        )
        node = cfg_to_engine_node(cfg)
        compare_with_reference(
            f"{project_name}_{mode.name.lower()}_global_{camel_to_snake(cb_info.name)}_{passes}_nodes",
            format_engine_node(node),
        )


def camel_to_snake(name: str) -> str:
    return "".join(f"_{c.lower()}" if c.isupper() else c for c in name).lstrip("_")



from math import floor

from pydori.lib.layer import LAYER_LANE, LAYER_PREVIEW_COVER, LAYER_STAGE, LAYER_TIME_LINE, get_z
from pydori.lib.layout import END_LANE, START_LANE
from pydori.lib.skin import Skin
from pydori.lib.ui import init_ui
from pydori.preview.layout import (
    PREVIEW_BAR_LINE_ALPHA,
    PREVIEW_COVER_ALPHA,
    PREVIEW_Y_MAX,
    PREVIEW_Y_MIN,
    PreviewData,
    PreviewLayout,
    init_preview_layout,
    layout_preview_bar_line,
    layout_preview_lane,
    layout_preview_stage_border_left,
    layout_preview_stage_border_right,
    print_at_time,
)
from sonolus.script.archetype import PreviewArchetype, callback
from sonolus.script.printing import PrintColor, PrintFormat
from sonolus.script.quad import Rect
from sonolus.script.runtime import screen


class PreviewStage(PreviewArchetype):
    """Draws the stage and time bar lines."""

    name = "Stage"

    @callback(order=1)
    def preprocess(self):
        init_ui()
        init_preview_layout()

    def render(self):
        self.render_border()
        self.render_lanes()
        self.render_cover()
        self.render_times()

    def render_border(self):
        for col in range(PreviewLayout.column_count):
            left_border_layout = layout_preview_stage_border_left(col)
            right_border_layout = layout_preview_stage_border_right(col)
            Skin.stage_left_border.draw(left_border_layout, z=LAYER_STAGE)
            Skin.stage_right_border.draw(right_border_layout, z=LAYER_STAGE)

    def render_lanes(self):
        for col in range(PreviewLayout.column_count):
            for lane in range(START_LANE, END_LANE + 1):
                layout = layout_preview_lane(lane, col)
                Skin.lane.draw(layout, z=LAYER_LANE)

    def render_cover(self):
        left_x = screen().l
        right_x = PreviewLayout.column_width * PreviewLayout.column_count + 1
        Skin.cover.draw(
            Rect(
                l=left_x,
                r=right_x,
                b=-1,
                t=PREVIEW_Y_MIN,
            ),
            z=get_z(LAYER_PREVIEW_COVER),
            a=PREVIEW_COVER_ALPHA,
        )
        Skin.cover.draw(
            Rect(
                l=left_x,
                r=right_x,
                b=PREVIEW_Y_MAX,
                t=1,
            ),
            z=get_z(LAYER_PREVIEW_COVER),
            a=PREVIEW_COVER_ALPHA,
        )

    def render_times(self):
        for time in range(floor(PreviewData.last_time) + 1):
            print_at_time(
                time,
                time,
                fmt=PrintFormat.TIME,
                decimal_places=0,
                color=PrintColor.CYAN,
                side="left",
            )
            Skin.time_line.draw(
                layout_preview_bar_line(time, "left_only"),
                z=LAYER_TIME_LINE,
                a=PREVIEW_BAR_LINE_ALPHA,
            )



from sonolus.script.particle import StandardParticle, particles


@particles
class Particles:
    lane: StandardParticle.LANE_LINEAR

    tap_linear: StandardParticle.NOTE_LINEAR_TAP_CYAN
    tap_circular: StandardParticle.NOTE_CIRCULAR_TAP_CYAN

    hold_linear: StandardParticle.NOTE_LINEAR_TAP_GREEN
    hold_circular: StandardParticle.NOTE_CIRCULAR_TAP_GREEN
    hold_active_circular: StandardParticle.NOTE_CIRCULAR_HOLD_GREEN

    flick_linear: StandardParticle.NOTE_LINEAR_ALTERNATIVE_RED
    flick_circular: StandardParticle.NOTE_CIRCULAR_ALTERNATIVE_RED

    right_flick_linear: StandardParticle.NOTE_LINEAR_ALTERNATIVE_YELLOW
    right_flick_circular: StandardParticle.NOTE_CIRCULAR_ALTERNATIVE_YELLOW

    left_flick_linear: StandardParticle.NOTE_LINEAR_ALTERNATIVE_PURPLE
    left_flick_circular: StandardParticle.NOTE_CIRCULAR_ALTERNATIVE_PURPLE
